local LrLogger = import 'LrLogger'
local logger = LrLogger("ttg")

local LrDate = import 'LrDate'
local LrPathUtils = import 'LrPathUtils'
local LrStringUtils = import 'LrStringUtils'
local LrFunctionContext = import 'LrFunctionContext'
local LrTasks = import 'LrTasks'

local writeTimeout = { timeout=60, asynchronous=true }

MetadataHandler = {}
	
-- =============================================================================================================================================

local EMPTY_MARKER = "\14"


--TODO: implement properly
local function scalar(v)
	return v
end


local function strip_mm(str)
   if str then
      return scalar(str:gsub(' mm$', ''))
   else
      return str
   end
end

local function F(filename)
   if filename then
      return LrPathUtils.removeExtension(filename)
   else
      return "filename"
   end
end



-- Re-implemented by Ben =======================================================================================================

local function notTable(t)
	if type(t) ~= 'table' then
		return true
	else
		return false
	end
end

local function isTable(t)
	if type(t) == 'table' then
		return true
	else
		return false
	end
end

local function isString(s)
	if type(s) == 'string' then
		return true
	else
		return false
	end
end

local function notNumber(v)
	if v == nil and v == '' or type(v) ~= 'number' then
		return true
	else
		return false
	end
end

local function table_simplecopy(orig)
	local orig_type = type(orig)
	local copy
	if orig_type == 'table' then
		copy = {}
		for orig_key, orig_value in pairs(orig) do
		   copy[orig_key] = orig_value
		end
	else -- number, string, boolean, etc
		copy = orig
	end
	
	return copy
end

local function LrPhoto_path(photo)
	return photo.path
end

local function LrPhoto_uuid(photo)
	return photo:getRawMetadata('uuid')
end

--[[
--TODO: check with Jeffrey

local function read_catalog(f)
	local catalog = (import 'LrApplication').activeCatalog()
	catalog:withWriteAccessDo( "updateUrl", f 
end
]]--

--TODO: check with Jeffrey
local function gsub_until_unchanged(result, chars, char)

	--return result
	return string.gsub(result, chars, char)
end

local function split(line, delim, maxNb)
	local found = {}
	local maxNb = maxNb or 0

	if string.find(line, delim, start) == nil then
		return { line }
	end

	local pat = "(.-)" .. delim .. "()"
	local nb = 0
	local lastPos
	for part, pos in string.gfind(line, pat, start) do
		nb = nb + 1
		found[nb] = part
		lastPos = pos
	end
	-- Handle the last field
	if nb ~= maxNb then
		found[nb + 1] = string.sub(line, lastPos)
	end

	--logger:debug('result: '..line)
	--logger:debug('found: '..TTGUtils.dumpvar(found))

	return found
end

local function ValidateState(state)
end

local function DumpLog(t)
end

local function Assert(v, message)
	assert(v)
end

local function Logf()
end

local sprintf = string.format


-- End Re-implemented by Ben =======================================================================================================










local IsCropped  = LOC("$$$/xxx=cropped")
local NotCropped = LOC("$$$/xxx=uncropped")

local FlashDidFire = LOC("$$$/xxx=flash fired")
local FlashNotFire = LOC("$$$/xxx=no flash")
local FlashUnknown = LOC("$$$/xxx=unknown flash")



local function String(item)
   if item == nil then
      return ""
   else
      return tostring(item)
   end
end

local FixupPath
if WIN_ENV then
   function FixupPath(val)
      return scalar(val:gsub('\\', ' > '))
   end
else
   function FixupPath(val)
      return scalar(val:gsub('/', ' > '))
   end
end

local function GrabLeadingFloat(text)
   local T, E = (text:match('^%d+%.%d*(%D*)$')
              or
              text:match('^%.%d+(%D*)$')
              or
              text:match('^%d+(%D*)$'))
   if T then
      return tonumber(T)
   else
      return nil
   end
end

local function FixApertureString(text)
   return scalar(String(text):gsub(LOC("$$$/xxx=^U+0192 / "), "f/"))
end

local function get_photo_time(x)
   return x.photo:getRawMetadata('dateTimeOriginal')
       or x.photo:getRawMetadata('dateTimeDigitized')
       or x.photo:getRawMetadata('dateTime')
end

local function OnlineUrlHelp(name)
   return LOC("$$$/xxx=The url of the image at ^1 (only if previously uploaded with my Upload-to-^1 plugin).", name)
end

local function OnlineUrlGet(pluginId, fieldName)
   return function(x)
             -- we just ignore errors
             local success, URL = LrTasks.pcall(function()
                                                   return x.photo:getPropertyForPlugin(pluginId, fieldName, nil, true)
                                                end)
             if success and URL then
                if x.arg then
                   return sprintf("<a href='%s'>%s</a>", URL, x.arg)
                else
                   return URL
                end
             else
                return ""
             end
          end
end

local function photo_time_component(x, format)
   local time = get_photo_time(x)
   if not time then
      return ""
   else
      local val = LrDate.timeToUserFormat(time, format)
      Logf("time is %s [%s] for %s, %s = [%s]\n", time, x.photo:getFormattedMetadata('dateTimeOriginal'), tostring(x.photo), format, val)
      return val
   end
end

local function IfKeyword(x, metadataKey, metadataValue)
   local arg = x.arg or ""
   local wanted_keyword, if_have, if_not = unpack(split(arg, ";", 3))

   if not wanted_keyword:match("%S") then
      return nil, LOC("$$$/xxx=missing keyword for {IfKeyword} token")
   end

   wanted_keyword = wanted_keyword:gsub('^%s+', ''):gsub('%s+$', '')

   local keywords_list = metadataValue or x.photo:getFormattedMetadata(metadataKey)

   local keywords = split(keywords_list, ", ")
   for _, keyword in ipairs(keywords) do
      if keyword == wanted_keyword then
         -- if no IF text given, use the keyword itself
         return if_have == "" and wanted_keyword or if_have
      end
   end
   return if_not or ""
end

local function DaysSince(StartingSpec, timeStamp, multiplier, delta)
   local YYYY, MM, DD, hh, mm
   if StartingSpec then
      YYYY, MM, DD, hh, mm = StartingSpec:match('^%s*(%d%d%d%d)%-(%d%d?)%-(%d%d?)%s+(%d%d?):(%d%d)%s*$')
   end
   if not YYYY and StartingSpec then
      YYYY, MM, DD, hh, mm = StartingSpec:match('^%s*(%d%d%d%d)%-(%d%d?)%-(%d%d?)%s*$')
      hh = 0
      mm = 0
   end
   if not YYYY then
      return LOC("$$$/xxx=Unknown date format (want YYYY-MM-DD with optional HH:MM)")
   end

   local Then = LrDate.timeFromComponents(YYYY, MM, DD, hh, mm, 0, true) -- timezeon of true means local TZ
   local val = tostring(math.floor((delta or 0) + (multiplier or 1) * (timeStamp - Then)/(24 * 3600)))

   return val
end

local Tokens = {
   --KEY(_RAW_TEXT_)
   _RAW_TEXT_ = {
      NotForLuaEmbedding = true,
      need_x = 'arg',
      get = function(x) return x.arg end, -- support for the "text" pseudo-token
      no_show = true
   },

   --KEY(NOJOINERS)
   NOJOINERS = {
      NotForLuaEmbedding = true,
      get = function(x) return "\6" end,
      no_show = true
   },

   --KEY(YYYY)
   YYYY = {
      need_x = 'photo', --BAREOK(photo)
      get = function(x) return photo_time_component(x, '%Y') end,
      help = LOC("$$$/x1361=The photo year as a four-digit string."),
   },

   --KEY(YY)
   YY = {
      need_x = 'photo',
      get = function(x) return photo_time_component(x, '%y') end,
      help = LOC("$$$/x1362=The photo year as a two-digit string."),
   },

   --KEY(MM)
   MM = {
      need_x = 'photo',
      get = function(x) return photo_time_component(x, '%m') end,
      help = LOC("$$$/x1363=The photo month as a two-digit string."),
   },

   --KEY(DD)
   DD = {
      need_x = 'photo',
      get = function(x) return photo_time_component(x, '%d') end,
      help = LOC("$$$/x1364=The photo date-of-month as a two-digit string."),
   },

   --KEY(HH)
   HH = {
      need_x = 'photo',
      get = function(x) return photo_time_component(x, '%H') end,
      help = LOC("$$$/x1364=The photo hour in 24-hour notation, as a two-digit string."),
   },
   --KEY(HH12)
   HH12 = {
      need_x = 'photo',
      get = function(x) return scalar(photo_time_component(x, '%I'):gsub("^0", "")) end,
      help = LOC("$$$/x1364=The photo hour in 12-hour notation, as a one- or two-digit string."),
   },

   --KEY(MIN)
   MIN = {
      need_x = 'photo',
      get = function(x) return photo_time_component(x, '%M') end,
      help = LOC("$$$/x1364=The photo minute as a two-digit string."),
   },

   --KEY(SS)
   SS = {
      need_x = 'photo',
      get = function(x) return photo_time_component(x, '%S') end,
      help = LOC("$$$/x1365=The photo seconds value as a two-digit string."),
   },

   --KEY(AMPM)
   AMPM = {
      need_x = 'photo',
      get = function(x) return scalar(photo_time_component(x, '%p'):gsub("^ ", "")) end,
      help = LOC("$$$/x1365=AM/PM designation for the photo time"),
   },

   --KEY(Mon)
   Mon = {
      need_x = 'photo',
      get = function(x) return photo_time_component(x, '%b') end,
      help = LOC("$$$/x1366=The photo month as a three-character string."),
   },

   --KEY(Month)
   Month = {
      need_x = 'photo',
      get = function(x) return photo_time_component(x, '%B') end,
      help = LOC("$$$/x1367=The photo month as a string."),
   },

   --KEY(Date)
   Date = {
      need_x = 'photo',
      get = function(x) return photo_time_component(x, '%e') end,
      help = LOC("$$$/x1368=The photo date-of-month as a one- or two-digit string."),
   },

   --KEY(yyyy)
   yyyy = {
      for_current_time = true,
      get = function(x) return LrDate.timeToUserFormat(LrDate.currentTime(), '%Y') end,
      help = LOC("$$$/x1369=The current year as a four-digit string."),
   },

   --KEY(yy)
   yy = {
      for_current_time = true,
      get = function(x) return LrDate.timeToUserFormat(LrDate.currentTime(), '%y') end,
      help = LOC("$$$/x1370=The current year as a two-digit string."),
   },

   --KEY(mm)
   mm = {
      for_current_time = true,
      get = function(x) return LrDate.timeToUserFormat(LrDate.currentTime(), '%m') end,
      help = LOC("$$$/x1371=The current month as a two-digit string."),
   },

   --KEY(dd)
   dd = {
      for_current_time = true,
      get = function(x) return LrDate.timeToUserFormat(LrDate.currentTime(), '%d') end,
      help = LOC("$$$/x1372=The current date-of-month as a two-digit string."),
   },

   --KEY(hh)
   hh = {
      for_current_time = true,
      get = function(x) return LrDate.timeToUserFormat(LrDate.currentTime(), '%H') end,
      help = LOC("$$$/x1371=The current time's hour in 24-hour notation, as a two-digit string."),
   },
   --KEY(hh12)
   hh12 = {
      for_current_time = true,
      get = function(x) return scalar(LrDate.timeToUserFormat(LrDate.currentTime(), '%I'):gsub("^0", "")) end,
      help = LOC("$$$/x1371=The current time's hour in 12-hour notation, as a one- or two-digit string."),
   },
   --KEY(min)
   min = {
      for_current_time = true,
      get = function(x) return LrDate.timeToUserFormat(LrDate.currentTime(), '%M') end,
      help = LOC("$$$/x1372=The current time's minutes as a two-digit string."),
   },
   --KEY(ss)
   ss = {
      for_current_time = true,
      get = function(x) return LrDate.timeToUserFormat(LrDate.currentTime(), '%S') end,
      help = LOC("$$$/x1372=The current time's seconds as a two-digit string."),
   },
   --KEY(ampm)
   ampm = {
      for_current_time = true,
      get = function(x) return scalar(LrDate.timeToUserFormat(LrDate.currentTime(), '%p'):gsub("^ ", "")) end,
      help = LOC("$$$/x1372=The current time's AM/PM designation."),
   },

   --KEY(mon)
   mon = {
      for_current_time = true,
      get = function(x) return LrDate.timeToUserFormat(LrDate.currentTime(), '%b') end,
      help = LOC("$$$/x1373=The current month as a three-character string."),
   },
   --KEY(month)
   month = {
      for_current_time = true,
      get = function(x) return LrDate.timeToUserFormat(LrDate.currentTime(), '%B') end,
      help = LOC("$$$/x1374=The current month as a string."),
   },

   --KEY(date)
   date = {
      for_current_time = true,
      get = function(x) return LrDate.timeToUserFormat(LrDate.currentTime(), '%e') end,
      help = LOC("$$$/x1375=The current date-of-month as a one- or two-digit string."),
   },

   --KEY(Caption)
   Caption = {
      need_x = 'photo',
      get    = function(x) return x.photo:getFormattedMetadata('caption') end,
      help   = LOC("$$$/xxx=The photo's ^[Caption^] metadata item in Lightroom's database."),
   },

   --KEY(CopyName)
   CopyName = {
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('copyName') end,
      help = LOC("$$$/x1376=Name of the copy (master or virtual) being exported, if any."),
   },

   --KEY(DaysSince)
   DaysSince = {
      NotForLuaEmbedding = true,
      need_x = 'arg',
      get = function(x)
               return DaysSince(x.arg, LrDate.currentTime(), 1, 0)
            end,
      sample = LrDate.timeToUserFormat(LrDate.currentTime() - 360 * 24 * 3600, "{DaysSince=%Y-12-31}"),
      help = LOC("$$$/xxx=Days since the date or date/time in the argument.")
   },

   --KEY(DaysUntil)
   DaysUntil = {
      NotForLuaEmbedding = true,
      need_x = 'arg',
      get = function(x)
               return DaysSince(x.arg, LrDate.currentTime(), -1, 0)
            end,
      sample = LrDate.timeToUserFormat(LrDate.currentTime(), "{DaysUntil=%Y-12-31}"),
      help = LOC("$$$/xxx=Days until the date or date/time in the argument.")
   },

   --KEY(PhotoDaysSince)
   PhotoDaysSince = {
      NotForLuaEmbedding = true,
      need_x = 'photo',
      get = function(x)
               local time = get_photo_time(x)
               if not time then
                  return ""
               else
                  return x.arg and DaysSince(x.arg, time, 1, 0)
               end
            end,
      sample = LrDate.timeToUserFormat(LrDate.currentTime() - 360 * 24 * 3600, "{PhotoDaysSince=%Y-12-31}"),
      help = LOC("$$$/xxx=Like {DaysSince}, but references the photo's date instead of the current date.")
   },


   --KEY(PhotoDaysUntil)
   PhotoDaysUntil = {
      NotForLuaEmbedding = true,
      need_x = 'photo',
      get = function(x)
               local time = get_photo_time(x)
               if not time then
                  return ""
               else
                  return x.arg and DaysSince(x.arg, time, -1, 0)
               end
            end,
      sample = LrDate.timeToUserFormat(LrDate.currentTime(), "{PhotoDaysUntil=%Y-12-31}"),
      help = LOC("$$$/xxx=Like {DaysUntil}, but references the photo's date instead of the current date.")
   },


   --KEY(Empty)
   Empty = {
      get = function(x) return EMPTY_MARKER end,
      help = LOC("$$$/x1377=An empty string, perhaps useful for testing."),
   },

   --KEY(Filename)
   Filename = {
      get = function(x) return F(x.Filename or (x.photo and x.photo:getFormattedMetadata('fileName'))) end,
      help = LOC("$$$/x1378=The filename (without path and without extension) of the exported copy."),
   },

   --KEY(LibraryFilename)
   LibraryFilename = {
      need_x = 'photo',
      get = function(x) return F(x.photo:getFormattedMetadata('fileName')) end,
      help = LOC("$$$/x1379=The filename (without path and without extension) of the master photo in the catalog."),
   },

   --KEY(LIBRARYFILENAME)
   LIBRARYFILENAME = {
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('fileName')                  end,
      help = LOC("$$$/x1380=Like {LibraryFilename}, but includes the extension."),
   },

   --KEY(FILENAME)
   FILENAME = {
      get = function(x) return x.Filename or (x.photo and x.photo:getFormattedMetadata('fileName'))   end,
      help = LOC("$$$/x1381=Like {Filename}, but includes the extension."),
   },

   --KEY(NumericFilenameSuffix)
   NumericFilenameSuffix = {
      need_x = 'photo',
      get = function(x)
         filename = F(x.Filename or (x.photo and x.photo:getFormattedMetadata('fileName')))
         return string.match(filename, '[1-9][0-9]*$')
      end,
   },

   --KEY(NUMERICFILENAMESUFFIX)
   NUMERICFILENAMESUFFIX = {
      need_x = 'photo',
      get = function(x)
         filename = F(x.Filename or (x.photo and x.photo:getFormattedMetadata('fileName')))
         return string.match(filename, '[0-9]+$')
      end,
   },

   --KEY(NumericLibraryFilenameSuffix)
   NumericLibraryFilenameSuffix = {
      need_x = 'photo',
      get = function(x)
         filename = F(x.photo:getFormattedMetadata('fileName'))
         return string.match(filename, '[1-9][0-9]*$')
      end,
   },

   --KEY(NUMERICLIBRARYFILENAMESUFFIX)
   NUMERICLIBRARYFILENAMESUFFIX = {
      need_x = 'photo',
      get = function(x)
         filename = F(x.photo:getFormattedMetadata('fileName'))
         return string.match(filename, '[0-9]+$')
      end,
   },

   --KEY(Filetype)
   Filetype = {
      need_x = 'photo',
      get = function(x)
               local type = x.photo:getRawMetadata('fileFormat')
               -- Earlier I used 'fileType', which returns 'JPEG' instead of 'JPG', and 'Video' instead of 'VIDEO',
               -- so map those accordingly.
               --
               -- Also, DNG may have "-LOSSLESS/LOSSY/REDUCED" appended, and others may have "-CMYK" appended,
               -- so we want to get rid of that.
               --
               if type:match('^JPG') then
                  return 'JPEG'
               elseif type:match('^VIDEO') then
                  return 'Video'
               else
                  return scalar(type:gsub('%-.*$', ''))
               end
            end, -- one of: 'JPEG', 'RAW', 'DNG', 'TIFF', 'PSD', or 'Video'.
      help = LOC("$$$/x1382=One of: ^[JPEG^], ^[RAW^], ^[DNG^], ^[TIFF^], ^[PSD^], or ^[Video^]."),
   },

   --KEY(FocalLength)
   FocalLength = {
      need_x = 'photo',
      get = function(x) return strip_mm(x.photo:getFormattedMetadata('focalLength')) end,
      help = LOC("$$$/x1383=Focal Length as a raw number."),
   },

   --KEY(FocalLength35)
   FocalLength35 = {
      need_x = 'photo',
      get = function(x) return strip_mm(x.photo:getFormattedMetadata('focalLength35mm')) end,
      help = LOC("$$$/x1384=Focal Length (in 35mm format) as a raw number."),
   },

   --KEY(FocalLengthMM)
   FocalLengthMM = {
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('focalLength')               end,
      help = LOC("$$$/x1385=Focal Length with ^[mm^] appended."),
   },

   --KEY(FocalLength35MM)
   FocalLength35MM = {
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('focalLength35mm')           end,
      help = LOC("$$$/x1386=Focal Length (in 35mm format) with ^[mm^] appended."),
   },

   --KEY(Folder)
   Folder = {
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('folderName')                end,
      help = LOC("$$$/x1387=The name of the folder (without path) in which the the master photo in the catalog resides."),
   },

   --KEY(Path)
   Path = {
      help = LOC("$$$/xxx=The path of the folder holding the master photo on disk."),
      need_x = 'photo',
      get = function(x)
               return LrPathUtils.parent(LrPhoto_path(x.photo))
            end,
      forAutoDestination = FixupPath,
   },

   --KEY(FullExportedFile)
   FullExportedFile = {
      get  = function(x) return x.FullExportedFile end,
      help = LOC("$$$/xxx=The full filename (with path and extension) of the exported copy."),
   },

   --KEY(FullExportedFolder)
   FullExportedFolder = {
      get  = function(x)
                if x.FullExportedFile then
                   return LrPathUtils.parent(x.FullExportedFile)
                end
             end,
      help = LOC("$$$/xxx=The full path to the folder holding the exported copy."),
   },


   --KEY(FullMasterFile)
   FullMasterFile = {
      help = LOC("$$$/xxx=The full path and filename of the master photo on disk."),
      need_x = 'photo',
      get = function(x)
               return LrPhoto_path(x.photo)
            end,
      forAutoDestination = FixupPath,
   },

   --KEY(FullMasterFolder)
   FullMasterFolder = {
      help = LOC("$$$/xxx=The full path and filename of the master photo on disk."),
      need_x = 'photo',
      get = function(x)
               return LrPathUtils.parent(LrPhoto_path(x.photo))
            end,
      forAutoDestination = FixupPath,
   },

   --KEY(Headline)
   Headline = {
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('headline')                  end,
      help = LOC("$$$/xxx=The photo's ^[Headline^] metadata item in Lightroom's database."),
   },

   --KEY(Rating)
   Rating = {
      help = LOC("$$$/x1389=Number of stars, ^[0^] through ^[5^], or, in the form {Rating=*}, a string of zero through five asterisks."),
      need_x = 'photo',
      get = function(x)
               local val = x.photo:getRawMetadata('rating')
               if not val then
                  val = 0
               end
               if not x.arg then
                  return tostring(val)
               end

               local Text = ""
               while val > 0 do
                  Text = Text .. x.arg
                  val = val - 1
               end
               return Text
            end,
   },

   --KEY(Title)
   Title = {
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('title')                     end,
      help = LOC("$$$/xxx=The photo's ^[Title^] metadata item in Lightroom's database."),
   },

   --KEY(Copyright)
   Copyright       = {
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('copyright')                 end,
      help = LOC("$$$/xxx=The photo's ^[Copyright^] metadata item in Lightroom's database."),
   },

   --KEY(OriginalWidth)
   OriginalWidth = {
      need_x = 'photo',
      get = function(x) return String(x.photo:getRawMetadata('dimensions').width)      end, -- the width of the ORIGINAL IMAGE (not the exported one)
      help = LOC("$$$/x1390=The original (prior to any crop) width of the master image in the library; may differ from the library or exported width."),
   },

   --KEY(OriginalHeight)
   OriginalHeight = {
      need_x = 'photo',
      get = function(x) return String(x.photo:getRawMetadata('dimensions').height)     end, -- the height of the ORIGINAL IMAGE (not the exported one)
      help = LOC("$$$/x1391=The original (prior to any crop)  height of the master image in the library; may differ from the library or exported height."),
   },

   --KEY(CroppedWidth)
   CroppedWidth = {
      need_x = 'photo',
      get = function(x) return String(x.photo:getRawMetadata('croppedDimensions').width)      end, -- the width of the ORIGINAL IMAGE (not the exported one)
      help = LOC("$$$/x1390=The (possibly post crop) width of the image in the library; may differ from the master or exported width."),
   },

   --KEY(CroppedHeight)
   CroppedHeight = {
      need_x = 'photo',
      get = function(x) return String(x.photo:getRawMetadata('croppedDimensions').height)     end, -- the height of the ORIGINAL IMAGE (not the exported one)
      help = LOC("$$$/x1391=The (possibly post crop) height of the image in the library; may differ from the master or exported height."),
   },

   --KEY(UUID)
   UUID = {
      need_x = 'photo',
      get = function(x)
               return LrPhoto_uuid(x.photo)
            end,

      help = LOC("$$$/x1392=Lightroom essentially-random unique id for the image."),
   },

   --KEY(-)
   ['-'] = {
      get = function(x) return "-" end, -- a raw hyphen that is never auto-removed
      help  = LOC("$$$/x1403=a raw hyphen that is never auto-removed."),
   },

   --KEY(,)
   [','] = {
      get = function(x) return "," end, -- a raw comma  that is never auto-removed
      help  = LOC("$$$/x1405=a raw comma that is never auto-removed."),
   },

   --KEY(Flash)
   Flash             = {
      need_x = 'photo',
      get = function(x)
               local val = x.photo:getRawMetadata('flash')
               if val == true then
                  return FlashDidFire
               elseif val == false then
                  return FlashNotFire
               else
                  return FlashUnknown
               end
            end,

      help = LOC("$$$/x1393=One of ^[^1^], ^[^2^], or ^[^3^].", FlashDidFire, FlashNotFire, FlashUnknown),
   },

   --KEY(Latitude)
   Latitude = {
      need_x = 'photo',
      get = function(x)
               local gps = x.photo:getRawMetadata('gps')
               return gps and String(gps.latitude) or ""
            end,
      help = LOC("$$$/x1394=Latitude, as a floating-point number. Use ^[Places^], e.g. {Latitude:Places=4}, to limit precision."),
   },

   --KEY(Longitude)
   Longitude = {
      need_x = 'photo',
      get = function(x)
               local gps = x.photo:getRawMetadata('gps')
               return gps and String(gps.longitude) or ""
            end,
      help = LOC("$$$/xxx=Longitude, as a floating-point number. Use ^[Places^], e.g. {Longitude:Places=4}, to limit precision."),
   },

   --KEY(ShutterSpeed)
   ShutterSpeed    = {
      need_x = 'photo',
      get = function(x)  return String(x.photo:getFormattedMetadata('shutterSpeed')) end,
      help = LOC("$$$/x1396=The shutter speed, formatted like ^[1/60 sec^]."),
   },

   --KEY(ShutterSpeedNum)
   ShutterSpeedNum = {
      need_x = 'photo',
      get = function(x)  return String(x.photo:getRawMetadata('shutterSpeed')) end,
      help = LOC("$$$/x1397=The shutter speed as a floating-point number of seconds; Use the ^[Places^] filter to limit precision."),
   },

   --KEY(SubjectDistance)
   SubjectDistance = {
      need_x = 'photo',
      get = function(x) return String(x.photo:getFormattedMetadata('subjectDistance')) end,
      help = LOC("$$$/x1398=The subject distance as reported by the lens, formatted like ^[1.2 m^]. Often wildly inaccurate."),
   },

   --KEY(SubjectDistanceNum)
   SubjectDistanceNum = {
      need_x = 'photo',
      get = function(x)
               return scalar(String(x.photo:getFormattedMetadata('subjectDistance')):gsub(' m$', ''))
            end,
      help = LOC("$$$/x1399=The subject distance as reported by the lens, as a floating-point number. Often wildly inaccurate."),
   },

   --KEY(EditCount)
   EditCount = {
      need_x = 'photo',
      get = function(x)  return String(x.photo:getRawMetadata('editCount')) end,
      help = LOC("$$$/x1400=A number that goes up each time a change is made in Lightroom to the image or its data."),
   },

   --KEY(Cropped)
   Cropped = {
      need_x = 'photo',
      get = function(x) return x.photo:getRawMetadata('isCropped') and IsCropped or NotCropped  end,
      help = LOC("$$$/x1401=Either ^[^1^] or ^[^2^].", IsCropped, NotCropped)
   },

   --KEY(OnlineID)
   OnlineID = {
      help = LOC("$$$/xxx=The ID of the image at the online service (Flickr only)."),
      get = function(x) --KEYS(OnlineID QuietOnAllErrors MissingLiveComponent)
               if x.OnlineID then
                  return tostring(x.OnlineID)
               elseif x.QuietOnAllErrors then
                  return nil
               else
                  LogNote("MissingLiveComponent: OnlineID")
                  x.MissingLiveComponent = true
                  return "XXX"
               end
            end,
   },

   --KEY(PluginProperty)
   PluginProperty = {
      need_x = 'photo', -- also need arg
      get = function(x)
               if not isString(x.arg) or x.arg:match("^%s*$") then
                  return nil, LOC("$$$/xxx=PluginPropery form is {PluginProperty=pluginid.fieldname}")
               end
               local pluginId, fieldName = x.arg:match("^(%S+)%.([^%s.]+)$")
               if not fieldName then
                  return nil, "Invalid argument in {PluginProperty} token"
               end

               local success, item = LrTasks.pcall(function()
                                                      return x.photo:getPropertyForPlugin(pluginId, fieldName, nil, true)
                                                   end)
               if success and item then
                  return tostring(item)
               else
                  return ""
               end
            end
   },


   --KEY(ExponoUrl)
   ExponoUrl = {
      help = OnlineUrlHelp('Expono'),
      need_x = 'photo',
      get = OnlineUrlGet('info.regex.lightroom.export.expono2', 'url')
   },

   --KEY(FacebookUrl)
   FacebookUrl = {
      help = OnlineUrlHelp('Facebook'),
      need_x = 'photo',
      get = OnlineUrlGet('info.regex.lightroom.export.facebook2', 'url')
   },

   --KEY(FlickrUrl)
   FlickrUrl = {
      help = OnlineUrlHelp('Flickr'),
      need_x = 'photo',
      get = OnlineUrlGet('info.regex.lightroom.export.flickr2', 'url')
   },

   --KEY(PhotobucketUrl)
   PhotobucketUrl = {
      help = OnlineUrlHelp('Photobucket'),
      need_x = 'photo',
      get = OnlineUrlGet('info.regex.lightroom.export.photobucket2', 'url')
   },

   --KEY(PicasawebUrl)
   PicasawebUrl = {
      help = OnlineUrlHelp('PicasaWeb'),
      need_x = 'photo',
      get = OnlineUrlGet('info.regex.lightroom.export.picasaweb2', 'url')
   },

   --KEY(SmugMugUrl)
   SmugMugUrl = {
      help = OnlineUrlHelp('SmugMug'),
      need_x = 'photo',
      get = OnlineUrlGet('info.regex.lightroom.export.smugmug2', 'url')
   },

   --KEY(ZenfolioUrl)
   ZenfolioUrl = {
      help = OnlineUrlHelp('Zenfolio'),
      need_x = 'photo',
      get = OnlineUrlGet('info.regex.lightroom.export.zenfolio2', 'url')
   },

   --KEY(ExposureManagerUrl)
   ExposureManagerUrl = {
      help = OnlineUrlHelp('ExposureManager'),
      need_x = 'photo',
      get = OnlineUrlGet('info.regex.lightroom.export.exposuremanager2', 'url')
   },


   --KEY(Aperture)
   Aperture = {
      help = LOC("$$$/xxx=The photo's aperture value, formatted like ^[f/4.5^]"),
      need_x = 'photo',
      get = function(x) return FixApertureString(x.photo:getFormattedMetadata('aperture')) end,
   },

   --KEY(ApertureNum)
   ApertureNum = {
      help = LOC("$$$/xxx=The photo's aperture value, as a raw number"),
      need_x = 'photo',
      get = function(x) return String(x.photo:getRawMetadata('aperture')) end,
   },

   --KEY(Artist)
   Artist = {
      help = LOC("$$$/xxx=The photo's ^[Artist^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('artist') end,
   },

   --KEY(CameraMake)
   CameraMake = {
      help = LOC("$$$/xxx=The photo's ^[Make^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('cameraMake') end,
   },

   --KEY(CameraModel)
   CameraModel = {
      help = LOC("$$$/xxx=The photo's ^[Model^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('cameraModel') end,
   },

   --KEY(CameraName)
   CameraName = {
      help = LOC("$$$/xxx=A sane version of the camera make/model combination"),
      need_x = 'photo',
      --get = function(x) return GetCameraName(x.photo) end,
      --get = function(x) return x.photo:getFormattedMetadata('cameraMake')..' '..x.photo:getFormattedMetadata('cameraModel') end,

		get = function(x) 
			local cameraMake = x.photo:getFormattedMetadata('cameraMake')
			local cameraModel = x.photo:getFormattedMetadata('cameraModel')

			if cameraMake ~= nil and cameraModel ~= nil then
				return cameraMake..' '..cameraModel
			elseif cameraMake ~= nil then
				return cameraModel
			elseif cameraMake ~= nil then
				return cameraModel
			else
				return ''
			end
		end,
   },

   --KEY(CameraSerialNumber)
   CameraSerialNumber = {
      help = LOC("$$$/xxx=The photo's ^[Serial Number^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('cameraSerialNumber') end,
   },

   --KEY(Category)
   Category = {
      help = LOC("$$$/xxx=The photo's ^[IPTC > Category^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('iptcCategory') end,
   },

   --KEY(City)
   City = {
      help = LOC("$$$/xxx=The photo's ^[Image > City^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('city') end,
   },

   --KEY(ColorLabel)
   ColorLabel = {
      help = LOC("$$$/xxx=The photo's color-label string."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('label') end,
   },

   --KEY(CopyrightUrl)
   CopyrightUrl = {
      help = LOC("$$$/xxx=The photo's ^[Copyright Info Url^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('copyrightInfoUrl') end,
   },

   --KEY(Country)
   Country = {
      help = LOC("$$$/xxx=The photo's ^[Image > Country^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('country') end,
   },

   --KEY(CountryCode)
   CountryCode = {
      help = LOC("$$$/xxx=The photo's ^[Image > ISO Country Code^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('isoCountryCode') end,
   },

   --KEY(Creator)
   Creator = {
      help = LOC("$$$/xxx=The photo's ^[Contact > Creator^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('creator') end,
   },

   --KEY(CreatorAddress)
   CreatorAddress = {
      help = LOC("$$$/xxx=The photo's ^[Contact > Address^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('creatorAddress') end,
   },

   --KEY(CreatorCity)
   CreatorCity = {
      help = LOC("$$$/xxx=The photo's ^[Contact > City^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('creatorCity') end,
   },

   --KEY(CreatorCountry)
   CreatorCountry = {
      help = LOC("$$$/xxx=The photo's ^[Contact > Country^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('creatorCountry') end,
   },

   --KEY(CreatorEmail)
   CreatorEmail = {
      help = LOC("$$$/xxx=The photo's ^[Contact > E-Mail^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('creatorEmail') end,
   },

   --KEY(CreatorJobTitle)
   CreatorJobTitle = {
      help = LOC("$$$/xxx=The photo's ^[Contact > Job Title^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('creatorJobTitle') end,
   },

   --KEY(CreatorPhone)
   CreatorPhone = {
      help = LOC("$$$/xxx=The photo's ^[Contact > Phone^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('creatorPhone') end,
   },

   --KEY(CreatorState)
   CreatorState = {
      help = LOC("$$$/xxx=The photo's ^[Contact > State / Province^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('creatorStateProvince') end,
   },

   --KEY(CreatorUrl)
   CreatorUrl = {
      help = LOC("$$$/xxx=The photo's ^[Contact > Website^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('creatorUrl') end,
   },

   --KEY(CreatorZip)
   CreatorZip = {
      help = LOC("$$$/xxx=The photo's ^[Contact > Postal Code^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('creatorPostalCode') end,
   },

   --KEY(DescriptionWriter)
   DescriptionWriter = {
      help = LOC("$$$/xxx=The photo's ^[IPTC > Description Writer^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('descriptionWriter') end,
   },

   --KEY(Exposure)
   Exposure = {
      help = LOC("$$$/xxx=The photo's exposure data, like ^[1/250 sec at f/5.0^]."),
      need_x = 'photo',
      get = function(x) return FixApertureString(x.photo:getFormattedMetadata('exposure')) end,
   },

   --KEY(ExposureBias)
   ExposureBias = {
      help = LOC("$$$/xxx=The photo's exposure-bias setting, like ^[-1 EV^]."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('exposureBias') end,
   },

   --KEY(ExposureBiasNum)
   ExposureBiasNum = {
      help = LOC("$$$/xxx=The photo's exposure-bias setting, as a raw number."),
      need_x = 'photo',
      get = function(x) return String(x.photo:getRawMetadata('exposureBias')) end,
   },

   --KEY(ExposureProgram)
   ExposureProgram = {
      help = LOC("$$$/xxx=The photo's exposure-program setting, like ^[Aperture priority^]."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('exposureProgram') end,
   },

   --KEY(Genre)
   Genre = {
      help = LOC("$$$/xxx=The photo's ^[Image > Intellectual Genre^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('intellectualGenre') end,
   },

   --KEY(GPSAltitude)
   GPSAltitude = {
      help = LOC("$$$/xxx=The photo's geoencoded altitude, e.g. ^[72.3 m^].  Also see {Altitude}. Use ^[Places^], e.g. {GPSAltitude:Places=4} to treat as numeric and to limit precision."),
      need_x = 'photo',
      get = function(x)
               return scalar(String(x.photo:getFormattedMetadata('gpsAltitude')):gsub('%.0', ' '))
            end,
   },

   --KEY(Altitude)
   Altitude = {
      help = LOC("$$$/xxx=The photo's geoencoded altitude as a number, in meters. Use ^[Places^], e.g. {Altitude:Places=1} to limit precision."),
      need_x = 'photo',
      get = function(x)
               return scalar(String(x.photo:getRawMetadata('gpsAltitude')):gsub('%.0', ''))
            end,
   },

   --KEY(GPSCoordinates)
   GPSCoordinates = {
      help = LOC("$$$/xxx=The photo's geoencoded coordinates, like ^[35^D1\'11.51\" N 135^D46\'16.05\" E^]"),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('gps') end,
   },

   --KEY(IfGeoencoded)
   IfGeoencoded = {
      NotForLuaEmbedding = true,
      help = LOC("$$$/xxx=Used as {IfGeoencoded=yes} or {IfGeoencoded=yes;no}, replaced by the text after '=' if geoencoded, or, optionally, the text after ';' if not."),
      sample = "{IfGeoencoded=abc;def}",
      need_x = 'photo',
      get = function(x)
               local is_geoencoded = x.photo:getFormattedMetadata('gps') or ""
               local arg = x.arg or ""
               local arg_if
               local arg_else
               if arg:match(";") then
                  arg_if, arg_else = arg:match("^([^;]*);(.*)")
               else
                  arg_if = arg
                  arg_else = ""
               end
               if is_geoencoded:match("%S") then
                  return arg_if
               else
                  return arg_else
               end
            end
   },

   --KEY(IfKeyword)
   IfKeyword = {
      NotForLuaEmbedding = true,
      help = LOC("$$$/xxx=Used as {IfKeyword=KEY;IF} or {IfKeyword=KEY;IF;ELSE}, replaced by the IF text if photo has keyword KEY, ELSE text (or nothing) if not."),
      sample = LOC("$$$/xxx={IfKeyword=priv;private;public}"),
      need_x = 'photo', -- I think
      get = function(x)
               return IfKeyword(x, 'keywordTags')
            end
   },
   --KEY(IfExportedKeyword)
   IfExportedKeyword = {
      help = LOC("$$$/xxx=Like {IfKeyword=...}, but considers only keywords marked for export."),
      sample = LOC("$$$/xxx={IfKeyword=priv;private;public}"),
      need_x = 'photo', -- I think
      get = function(x)
               return IfKeyword(x, 'keywordTagsForExport')
            end
   },

   --KEY(Instructions)
   Instructions = {
      help = LOC("$$$/xxx=The photo's ^[Workflow > Instructions^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('instructions') end,
   },

   --KEY(ISO)
   ISO = {
      help = LOC("$$$/xxx=The photo's ISO (sensor sensitivity) number."),
      need_x = 'photo',
      get = function(x) return String(x.photo:getRawMetadata('isoSpeedRating')) end,
   },

   --KEY(JobIdentifier)
   JobIdentifier = {
      help = LOC("$$$/xxx=The photo's ^[Workflow > Job Identifier^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('jobIdentifier') end,
   },

   --KEY(Lens)
   Lens = {
      help = LOC("$$$/xxx=The photo's lens information, like ^[24-70mm f/2.8^]."),
      need_x = 'photo',
      get = function(x)
               return scalar(String(x.photo:getFormattedMetadata('lens')):gsub('%.0', ''))
            end,
   },

   --KEY(Location)
   Location = {
      help = LOC("$$$/xxx=The photo's ^[Image > Location^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('location') end,
   },

   --KEY(Keywords)
   Keywords = {
      help = LOC("$$$/xxx=A comma-separated list of photo keywords that are marked for export."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('keywordTagsForExport') end,
   },

   --KEY(KeywordsAll)
   KeywordsAll = {
      help = LOC("$$$/xxx=A comma-separated list of all photo keywords (including those not marked for export)."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('keywordTags') end,
   },

   --KEY(MeteringMode)
   MeteringMode = {
      help = LOC("$$$/xxx=The photo's metering-mode setting, like ^[Pattern^]."),
      need_x = 'photo',
      get = function(x) return String(x.photo:getFormattedMetadata('meteringMode')) end,
   },

   --KEY(OtherCategories)
   OtherCategories = {
      help = LOC("$$$/xxx=The photo's ^[IPTC > Other Categories^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('iptcOtherCategories') end,
   },

   --KEY(Provider)
   Provider = {
      help = LOC("$$$/xxx=The photo's ^[Workflow > Provider^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('provider') end,
   },

   --KEY(Province)
   Province = {
      help = LOC("$$$/xxx=The photo's ^[Image > State / Province^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('stateProvince') end,
   },

   --KEY(RightsUsageTerms)
   RightsUsageTerms = {
      help = LOC("$$$/xxx=The photo's ^[Workflow > Rights Usage Terms^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('rightsUsageTerms') end,
   },

   --KEY(Scene)
   Scene = {
      help = LOC("$$$/xxx=The photo's ^[Image > Scene^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('scene') end,
   },

   --KEY(Software)
   Software = {
      help = LOC("$$$/xxx=The photo's ^[Software^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('software') end,
   },

   --KEY(Source)
   Source = {
      help = LOC("$$$/xxx=The photo's ^[Workflow > Source^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('source') end,
   },

   --KEY(State)
   State = {
      help = LOC("$$$/xxx=The photo's ^[Image > State / Province^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('stateProvince') end,
   },

   --KEY(SubjectCode)
   SubjectCode = {
      help = LOC("$$$/xxx=The photo's ^[IPTC > IPTC Subject Code^] metadata item in Lightroom's database."),
      need_x = 'photo',
      get = function(x) return x.photo:getFormattedMetadata('iptcSubjectCode') end,
   },

   --KEY(AspectRatio)
   AspectRatio = {
      need_x = 'photo',
      get = function(x)
               local arg = x.arg or "Portrait,Square,Landscape"
               local PSL = split(arg, "%,", 3)

               if x.photo then
               	--local size = x.photo:getRawMetadata('croppedDimensions')
               	local size = x.photo:getRawMetadata('dimensions')
                  if notTable(size) or notNumber(size.width) or notNumber(size.height) then
                     return "?"
                  elseif size.width > size.height then
                     -- landscape
                     return PSL[3] or ""
                  elseif size.width < size.height then
                     -- portrait
                     return PSL[1] or ""
                  else
                     -- square
                     return PSL[2] or ""
                  end
               end
               return "?"
            end,

      help = LOC("$$$/xxx=Returns ^[Portrait^}, ^[Square^], or ^[Landscape^]. See web page for customization."),
   },

}

if LrDate.formatMediumTime then

   -- KEY(d1)
   Tokens.d1 = {
      get = function(x) return LrDate.formatShortDate(LrDate.currentTime()) end,
      help = LOC("$$$/xxx=The current date in short system-local form")
   }
   -- KEY(d2)
   Tokens.d2 = {
      get = function(x) return LrDate.formatMediumDate(LrDate.currentTime()) end,
      help = LOC("$$$/xxx=The current date in medium system-local form")
   }
   -- KEY(d3)
   Tokens.d3 = {
      get = function(x) return LrDate.formatLongDate(LrDate.currentTime()) end,
      help = LOC("$$$/xxx=The current date in long system-local form")
   }
   -- KEY(t1)
   Tokens.t1 = {
      for_current_time = true,
      get = function(x) return LrDate.formatShortTime(LrDate.currentTime()) end,
      help = LOC("$$$/xxx=The current time in short system-local form")
   }
   -- KEY(t2)
   Tokens.t2 = {
      for_current_time = true,
      get = function(x) return LrDate.formatMediumTime(LrDate.currentTime()) end,
      help = LOC("$$$/xxx=The current time in long system-local form")
   }




   -- KEY(D1)
   Tokens.D1 = {
      need_x = 'photo',
      get = function(x)
               local time = get_photo_time(x)
               if not time then
                  return ""
               else
                  return LrDate.formatShortDate(time)
               end
            end,
      help = LOC("$$$/xxx=The photo date in short system-local form")
   }
   -- KEY(D2)
   Tokens.D2 = {
      need_x = 'photo',
      get = function(x)
               local time = get_photo_time(x)
               if not time then
                  return ""
               else
                  return LrDate.formatMediumDate(time)
               end
            end,
      help = LOC("$$$/xxx=The photo date in medium system-local form")
   }
   -- KEY(D3)
   Tokens.D3 = {
      need_x = 'photo',
      get = function(x)
               local time = get_photo_time(x)
               if not time then
                  return ""
               else
                  return LrDate.formatLongDate(time)
               end
            end,
      help = LOC("$$$/xxx=The photo date in long system-local form")
   }
   -- KEY(T1)
   Tokens.T1 = {
      need_x = 'photo',
      get = function(x)
               local time = get_photo_time(x)
               if not time then
                  return ""
               else
                  return LrDate.formatShortTime(time)
               end
            end,
      help = LOC("$$$/xxx=The photo time in short system-local form")
   }
   -- KEY(T2)
   Tokens.T2 = {
      need_x = 'photo',
      get = function(x)
               local time = get_photo_time(x)
               if not time then
                  return ""
               else
                  return LrDate.formatMediumTime(time)
               end
            end,
      help = LOC("$$$/xxx=The photo time in long system-local form")
   }

end

local function clean_up_lua_error(error)
   error = tostring(error or "<unknown>")
   local name = error:match(" global '([^']+)'") or error:match(" upvalue '([^']+)'") or error:match("undefined global: (%S+)")
   if name then
      return LOC("$$$/xxx=^[^1^] unknown", name)
   end

   return scalar(  error:gsub('^.*\"]:%d+:%s*', ''):gsub('"$', ''):gsub(" at line %d+", "")  )
end

-- KEY(LUA)
Tokens.LUA = {
   NotForLuaEmbedding = true,
   get = function(x)
            if not x.arg then
               return ""
            end
            local Lua_Arg = x.arg
            x.arg = nil

            local user_snippet = tostring(Lua_Arg)

            if not user_snippet:match("return") then
               user_snippet = "return " .. user_snippet
            end

            --DumpLog(user_snippet, "user_snippet")

            local KW  = { }
            local KWE = { }

            --
            -- No need to compute these if they don't seem to be referenced.
            -- We may get some false positives here, but that's okay.
            --
            if user_snippet:match("KW") then
               for _, keyword in ipairs(split(x.photo:getFormattedMetadata('keywordTags') or "", ',')) do
                  KW[keyword] = true
               end

               for _, keyword in ipairs(split(x.photo:getFormattedMetadata('keywordTagsForExport') or "", ',')) do
                  KWE[keyword] = true
               end
            end

            local my_string = table_simplecopy(string)
            my_string.dump = nil

            local my_math = table_simplecopy(math)
            my_math.randomseed = nil

            local my_os = {
               date     = os.date,       --KEY(date)
               tmpname  = os.tmpname,    --KEY(tmpname)
               time     = os.time,       --KEY(time)
               --difftime = os.difftime, --KEY(difftime)  --??
               --getenv   = os.getenv,   --KEY(getenv)    --??
            }

            local ENV = {
               --
               -- Built-ins to pass through
               --
               tonumber = tonumber,  --KEY(tonumber)
               tostring = tostring,  --KEY(tostring)
               type     = type,      --KEY(type)
               unpack   = unpack,    --KEY(unpack)
               pairs    = pairs,     --KEY(pairs)
               ipairs   = ipairs,    --KEY(ipairs)

               --
               -- Standard libraries to pass through
               --
               string   = my_string,                --KEY(string)
               math     = my_math,                  --KEY(math)
               os       = my_os,                    --KEY(os)
               table    = table_simplecopy(table),  --KEY(table)

               --
               -- Lr libraries
               --
               LrDate        = table_simplecopy(LrDate),         --KEY(LrDate)
               LrPathUtils   = table_simplecopy(LrPathUtils),    --KEY(LrPathUtils)
               LrStringUtils = table_simplecopy(LrStringUtils),  --KEY(LrStringUtils)

               --
               -- etc.
               --
               WIN     = WIN_ENV,        --KEY(WIN)
               MAC     = MAC_ENV,        --KEY(MAC)
               sprintf = string.format,  --KEY(sprintf)
               KW      = KW,             --KEY(KW)
               KWE     = KWE,            --KEY(KWE)

               photoTime = get_photo_time(x),      --KEY(photoTime)
               currentTime = LrDate.currentTime(), --KEY(currentTime)

               TBL     = function(...) return {...} end, -- empty table  --KEY(TBL)

               --
               -- "ne" is "non-empty"... Return nil if the item is nil or the empty string.
               --                        Otherwise, return the item.
               --
               ne      = function(item)  --KEY(ne)
                            if item == nil or tostring(item) == "" then
                               return nil
                            else
                               return item
                            end
                         end,

               --
               -- "nb" is "non-blank"... Return nil if the item is nil or a blank string (of any length)
               --                        Otherwise, return the item.
               --
               nb      = function(item)  --KEY(nb)
                            if item == nil or not tostring(item):match("%S") then
                               return nil
                            else
                               return item
                            end
                         end,

            }

            if LrFunctionContext.pcallWithEnvironment then
               function ENV.load(path) --KEY(load)
                  local result, error = ReadUserLuaFile(path, ENV)
                  DumpLog({
                             path    = path,
                             ENV     = tostring(ENV),
                             result  = result,
                             error   = error,
                          }, "returning from ENV.load()")
                  if not error then
                     return result
                  end
                  if x.QuietOnAllErrors then
                     return error
                  else
                     ThrowUserError("Error processing {LUA} token: " .. error)
                  end
               end
            end

            --            DumpLog({
            --               raw_croppedDimensions       = x.photo:getRawMetadata('croppedDimensions'),
            --               raw_dimensions              = x.photo:getRawMetadata('dimensions'),
            --               raw_aspectRatio             = x.photo:getRawMetadata('aspectRatio'),
            --               raw_width                   = x.photo:getRawMetadata('width'),
            --               raw_height                  = x.photo:getRawMetadata('height'),
            --               formatted_croppedDimensions = x.photo:getFormattedMetadata('croppedDimensions'),
            --               formatted_dimensions        = x.photo:getFormattedMetadata('dimensions'),
            --            }, "item size")



            -- LogNote("About to generate all tokens for Lua parse...")
            x.QuietOnAllErrors = true
            for name, rec in pairs(Tokens) do
               if rec.get and name:match("^%w") and not rec.NotForLuaEmbedding then
                  local val
                  if rec.need_x and not x[rec.need_x] then
                     -- don't know how to calculate this value with the data we have
                     val = "?"
                  else
                     val = rec.get(x) or ""
                  end
                  ENV[name] = val
               end
            end
            x.QuietOnAllErrors = false

            if x.ExtraLuaTokens then
               for name, val in pairs(x.ExtraLuaTokens) do
                  ENV[name] = val
               end
            end
            if x.ExtraTokens then
               for name, val in pairs(x.ExtraTokens) do
                  ENV[name] = val
               end
            end

            if LrFunctionContext.pcallWithEnvironment then
               local fn, error = loadstring(user_snippet)

               if not fn then
                  return LOC("$$$/xxx=Syntax Error: ^1", clean_up_lua_error(error))
               end

               local success, result = LrFunctionContext.pcallWithEnvironment(fn, ENV)

               if not success and result ~= nil then
                  -- result is an error message
                  return LOC("$$$/xxx=Runtime Error: ^1", clean_up_lua_error(tostring(result)))
               elseif result == nil then
                  return ""
               else
                  return tostring(result)
               end
            end




            local NAMES = { }
            local ARGS  = {}
            for name, value in pairs(ENV) do
               table.insert(NAMES, name)
               table.insert(ARGS,  value)
            end
            local namelist = table.concat(NAMES, ",")

            user_snippet = sprintf("do %s end    return ''", user_snippet)

            if not LrFunctionContext.pcallWithEmptyEnvironment then

               --
               -- Create local variables that hide all of our internal LR globals.
               -- Lua has a 200-variable-per-function limit, so to block all globals,
               -- we need to create a nested set to get full coverage....
               --
               local TOHIDE = {}
               local fcnnum = 0
               local function do_hide()
                  if #TOHIDE > 0 then
                     local tohide = table.concat(TOHIDE, ",")
                     local fcn_name = sprintf("__x_c_z_%d", fcnnum)
                     user_snippet = sprintf("\n\n\nfunction %s(%s)\n%s\nend\nreturn %s()\n\n", fcn_name, tohide, user_snippet, fcn_name)
                     TOHIDE = { }
                     fcnnum = fcnnum+1
                  end
               end

               for name in pairs(_G) do
                  if ENV[name] == nil then
                     table.insert(TOHIDE, name)
                     if #TOHIDE > 100 then
                        do_hide()
                     end
                  end
               end
               do_hide()

            end

            local lua = "return function(" .. namelist .. ") \n " .. user_snippet .. " end"

            --
            -- Dump it for debugging, but never dump for LR2 'cause it'll reveal all the globals.
            --
            --if LrFunctionContext.pcallWithEmptyEnvironment then
            --   DumpLog({ LUA = lua }, "LUA STRING")
            --end

            local fn, error = loadstring(lua)
            if not fn then
               return LOC("$$$/xxx=Syntax Error: ^1", clean_up_lua_error(error))
            end

            local func_to_be_protected = fn()

            local success, result
            if LrFunctionContext.pcallWithEmptyEnvironment then
               success, result = LrFunctionContext.pcallWithEmptyEnvironment(func_to_be_protected, unpack(ARGS))
            else
               func_to_be_protected = loadstring(string.dump(func_to_be_protected))
               success, result = LrTasks.pcall(func_to_be_protected, unpack(ARGS))
            end

            if not success and result ~= nil then
               -- result is an error message
               return LOC("$$$/xxx=Runtime Error: ^1", clean_up_lua_error(tostring(result)))
            elseif result == nil then
               return ""
            else
               return tostring(result)
            end
         end
}








function MetadataHandler.EnableOrdinalCounter() --KEY(EnableOrdinalCounter)
   --KEY(#)
   Tokens["#"] = {
      help = LOC("$$$/xxx=A counter. Use multiple ^[#^] to force a zero-padded number of at least that length."),
      get = function(x, etc)
               local num = x.ordinal or 1 --KEY(ordinal)
               if (etc.length or 1) > 1 then
                  num = sprintf("%0" .. tostring(etc.length) .. "d", num)
               else
                  num = tostring(num)
               end
               return num
            end
   }
end

local FILTER = {
   -- "Dash-Underscore to Space"
   --KEY(DU2S)
   DU2S = function(item)
             return scalar(item:gsub('[-_%s]+', ' '))
          end,

   -- "Dash to Space"
   --KEY(D2S)
   D2S = function(item)
            return scalar(item:gsub('[-%s]+', ' '))
          end,

   -- "Underscore to Space"
   --KEY(U2S)
   U2S = function(item)
            return scalar(item:gsub('[_%s]+', ' '))
         end,

   -- "Space to Underscore"
   --KEY(S2U)
   S2U = function(item)
            return scalar(item:gsub('[%s_]+', '_'))
         end,

   -- "Space to Dash"
   --KEY(S2D)
   S2D = function(item)
            return scalar(item:gsub('[%s-]+', '-'))
         end,


   --KEY(Places)
   Places = function(item, hash)
               --DumpLog { item = item, hash = hash }

               if not hash.arg then
                  return ''
               end
               local places = tostring(hash.arg)

               local fixup
               if places:match('^%d+$') then
                  fixup = function(num)
                             return sprintf("%." .. places .. "f", num)
                          end
               elseif places:match('^%-%d+$') then
                  local places_to_round = - tonumber(places)
                  local divisor = 10 ^ places_to_round
                  
                  fixup = function(num)
                             return tostring(math.floor(0.5 + num/divisor) * divisor)
                          end
               else
                  return LOC("$$$/xxx=(^[Places^] arg must be an integer)")
               end

               --
               -- Lua patterns require a lot of work to find a number (we'd want the regex
               -- /\d+(\.\d*)?|\.\d+/ ), so we just look for any run of [%d.] and if it converts to
               -- a number, that's great and we'll apply the formattting, and if not, we'll keep
               -- looking...
               --
               local converted = false
               local val = tostring(item):gsub('[%d.]+', function(snippet)
                                                            if converted then
                                                               return false -- just leave it
                                                            end
                                                            local as_num = tonumber(snippet)
                                                            if as_num == nil then
                                                               return false -- just leave it
                                                            end
                                                            converted = true
                                                            return fixup(as_num)
                                                         end)
               return val
            end,

   --KEY(Length)
   Length = function(item, hash)

               if not hash.arg then
                  return ''
               end
               if not tostring(hash.arg):match('^%d+$') then
                  return LOC("$$$/xxx=[^[Length^] arg must be a number]")
               end
               local length = tonumber(hash.arg)
               if length < 1 then
                  return LOC("$$$/xxx=[^[Length^] arg must be one or more]")
               end

               item = tostring(item)

               local Chars = utf8.bytes_to_characters(item)

               --Dump{ Chars = Chars, len = #Chars, length = length }

               if #Chars <= length then
                  return item
               end

               while #Chars > length do
                  table.remove(Chars)
               end


               return table.concat(Chars, '')
            end,
               
   --KEY(After)
   After = function(item, hash)
              if not hash.arg then
                 return ''
              end
              local i,j = item:lower():find(hash.arg:lower(), 1, true)
              if j then
                 local remain = item:sub(j + 1)
                 return scalar(remain:gsub('%s+$', ''))
              else
                 return ''
              end
           end,

   --KEY(UC)
   UC = function(item, hash)
           return LrStringUtils.upper(item)
        end,

   --KEY(LC)
   LC = function(item, hash)
           return LrStringUtils.lower(item)
        end,

   --KEY(UCFirst)
   UCFirst = function(item, hash)
                return utf8.ucFirst(item)
             end,

   --KEY(LCFirst)
   LCFirst = function(item, hash)
                return utf8.lcFirst(item)
             end,

   --KEY(UCWords)
   UCWords = function(item, hash)
                return utf8.ucWords(item)
             end,

}


--
-- Returns a string on error, a table on success
--
function MetadataHandler.parse(state, spec, x) --KEY(parse)

   ValidateState(state)

   local parts = {}
   local pos = 1
   while pos <= #spec do
      local A, B, C = spec:find('^([^{}]+)', pos)

      if A then
         table.insert(parts, { type = 'text', value = C })
         pos = B + 1
      else
         A, B, C = spec:find('^{([^{}]+)}', pos)
         if not A then
            if spec:find('^{', pos) then
               return LOC("$$$/xxx=Incomplete token ^[^1^]", spec:sub(pos))
            else
               return LOC("$$$/xxx=Error at ^[^1^]", spec:sub(pos))
            end
         end

         local list = { }
         for i, value in ipairs(split(C, '%s*|[%s|]*')) do 

            --
            -- Form is 'token:filter=arg', although the :filter and =arg parts are optional.
            -- Also, the form could be simply a string in quotes
            --
            local ORIG = value;

            local token
            local filter
            local arg
            local token_length

            -- If a raw-text string has been provided, inject the _RAW_TEXT_ token
            local raw_text = value:match('^"(.*)"$') or value:match("^'(.*)'$") 
            if raw_text then
               token = '_RAW_TEXT_'
               arg = raw_text
               filter = nil
            else
               value = value:gsub('^([#%w]+)', function(m) token  = m; return '' end)
               value = value:gsub('^:(%w+)',   function(m) filter = m; return '' end)
               value = value:gsub('^=(.*)',    function(m) arg    = m; return '' end)

               if not token then
                  return LOC("$$$/x54=Unknown token specification ^[{^1}^]", ORIG)
               end

               if filter and not FILTER[filter] then
                  return LOC("$$$/x55=Unknown filter ^[^1^] in ^[{^2}^]", filter, ORIG)
               end

               if not filter and value:match('^:') then
                  return LOC("$$$/x56=Missing filter name in ^[{^1}^]", ORIG)
               end

               if #value > 0 then
                  return LOC("$$$/x57=Unknown token specification at ^[^1^] in ^[{^2}^]", value, ORIG)
               end

               token_length = #token
               if token:match('^#+$') then
                  token = '#'
               end

               -- Dump({ ["old value"] = ORIG,
               --        ["new value"] = value,
               --        token = token,
               --        filter = filter,
               --        arg = arg })

               --DumpLog {
               --   x = x,
               --   token = token,
               --   XYZZY = true,
               --}

               if isTable(x) and isTable(x.ExtraTokens) and x.ExtraTokens[token] then
                  LogNote('using ExtraTokens[' .. tostring(token) .. "]")
                  -- we'll use this
					-- Allow tokens that were not found to be passed through
               elseif false and not Tokens[token] then
                  return LOC("$$$/x58=Unknown token ^[^1^]", token)
               end

               if type(Tokens[token]) == 'table' and Tokens[token].validate and not Tokens[token].validate(arg) then
                  return LOC("$$$/x59=Invalid argument ^[^1^] to token ^[^2^]", String(arg), token)
               end
            end

            local newrec = {
               token  = token,
               filter = filter,
               arg    = arg,
               length = token_length,
            }

            table.insert(list, newrec)
         end

         table.insert(parts, {
                         type = 'tokenlist',
                         value = list
                      })
         pos = B + 1
      end
   end

   return parts
end

function MetadataHandler.parses(state, spec, x) --KEY(parses)
   ValidateState(state)

   local parse = MetadataHandler.parse(state, spec, x)
   if notTable(parse) then
      return false, parse
   else
      return true
   end

end

--
-- In the specs, space, dash, and comma are considered "JOINING CHARACTERS"
-- and we treat instances of them specially when they're in the raw
-- specification as opposed to when they're part of a value brought in via
-- a token.
--
-- To allow us to differentiate between those that appear raw in the spec,
-- and those that are brought in, we map the former to unused characters
-- 0x01, 0x02, and 0x03 using these tables to map back and forth.
--
-- <br> and its friends are also mapped specially

local ReplacementTable = {
   ['-' ] = string.char(1),
   [' ' ] = string.char(2),
   [',' ] = string.char(3),
   ['\4'] = string.char(4), -- for <br> --KEY(\4)
   [':' ] = string.char(5), 
   -- \6 is a NOJOINERS token
   ['\n'] = string.char(7),  --KEY(\n)
   ['\r'] = string.char(8),  --KEY(\r)
   -- \14 is an EMPTY token (EMPTY_MARKER)

   -- NOTE: when adding items, make sure that they are not whitespace, or they'll get eaten by %s

}

--
-- Given a string of text, do all replacements necessary
--
local function DoReplacement(text)
   --return scalar(text:gsub('[-: ,\r\n]', ReplacementTable))
	return text
end

local function UnReplace(text)
   return scalar(text:gsub('\1+', '-'):gsub('\2+',' '):gsub('\3+',','):gsub('\5+',':'):gsub('\4+', '<br/>'):gsub('\7', "\n"):gsub('\8', "\r"):gsub(EMPTY_MARKER, ""))
end







--
-- returns (value, error)
--

-- x KEYS(isForAutoDestination photo arg preserveTrailing Filename)


function MetadataHandler.forPhoto(state, spec, x) --KEY(forPhoto)
   ValidateState(state)

	if spec == nil or spec == '' then
		return ''
	end

   local parse = MetadataHandler.parse(state, spec, x)
   if notTable(parse) then
      return nil, parse
   end

   local result = ''

   --[[if x and x.photo and LrMajorVersion <= 3 then
      TieShadowGPS(x.photo)
   end]]--

   local error = nil

	local catalog = (import 'LrApplication').activeCatalog()
   --read_catalog(function()
	catalog:withWriteAccessDo( "processMetadata", function( context ) 

	    for i, part in ipairs(parse) do
		-- for i=1,#parse do
		--   part = parse[i]

         if part.type == 'text' then

            --
            -- This is a section of raw text. Leading and trailing series of JOINING
            -- CHARACTERS are replaced with their holders. Embedded sequences are not
            -- touched because we assume that the user knows what they're doing.
            --
            -- <br> and friends, anywhere, are converted.
            --

            -- first handle <br> anywhere
            local to_add = part.value:gsub('<[bB][rR]/?>', '\4')

            -- now other stuff
            result = result .. to_add:gsub('^[-: ,\4\r\n]+', DoReplacement, 1):gsub('[-: ,\4\r\n]+$', DoReplacement, 1)

         elseif part.type == 'tokenlist' then

            -- we loop through the tokens, using only the first that has a non-empty value
            for j, token_hash in ipairs(part.value) do
			-- for j=1,#part.value do
			--   token_hash = part.value[j]

               local token_handle
               if isTable(x) and isTable(x.ExtraTokens) and x.ExtraTokens[token_hash.token] then
                  token_handle = function()
                                    return x.ExtraTokens[token_hash.token]
                                 end
               else
                  token_handle = Tokens[token_hash.token]
               end


               --Assert(token_handle, sprintf("missing handle for [%s]", token_hash.token))

               x.arg = token_hash.arg

               local val
               local extra

					if token_handle == val then
						-- Pass through tokens that were not matched
						val = '{'..token_hash.token..'}'
						--val = ''
               elseif type(token_handle) == 'function' then
                  if DEBUG then
                     DumpLog({
                                token_hash   = token_hash,
                                token_handle = "function",
                                x            = x
                             }, "calling token handle function")
                  end
                  val = token_handle(x)
               else
                  Assert(token_handle.get)
                  extra = token_handle
                  if DEBUG then
                      DumpLog({
                                 token_handle = token_handle,
                                 token_hash   = token_hash,
                                 x            = x,
                              }, "calling token_handle.x")
                   end

                  if token_handle.need_x and not x[token_handle.need_x] then
                     -- don't know how to calculate this value with the data we have
                     val, error = nil, nil
                  else
                     val, error = token_handle.get(x, token_hash)
                  end

                  if error then
                     DumpLog({ token_handle = token_handle, token_hash = token_hash, x = x, error = error }, "ERROR")
                     return
                  end

                  -- DumpLog( {
                  --     x = x,
                  --     token_hash = token_hash,
                  --     val = tostring(val)
                  --  }, "HERE")
               end


               if val then
                  val = tostring(val):gsub('^%s+', '',  1):gsub('%s+$', '', 1) -- strip leading/trailing whitespace

                  --
                  -- Apply filter it there's one and it's still valid
                  --
                  if token_hash.filter and FILTER[token_hash.filter] then
                     val = FILTER[token_hash.filter](val, token_hash)
                  end


                  if x.isForAutoDestination and val and isTable(extra) and extra.forAutoDestination then
                     val = extra.forAutoDestination(val)
                  end

                  if val ~= '' then
                     -- if anything's left, we'll use it.... 
                     result = result .. val
                     break
                  end
               end

            end
            -- nothing came of it, so we just add nothing.
         else
            Assert(false)
         end
      end
   end, writeTimeout)


   -- Added Sep 3, 2011 to report errors in token processing to the user...
   if error then
      return nil, error
   end

   --
   -- Now that we have the result, we may need to deal with the special JOINING CHARACTERS that we replaced.
   -- The special treatments are:
   --    * leading and trailing sequences are removed.
   --

   DumpLog(result, "Spec before")

   if not (x and x.preserveTrailing) then
      result = result:gsub('^[\1\2\3\4\5\7\8]+', '', 1):gsub('[\1\2\3\4\5\7\8]+$', '', 1)
   end

   --
   -- Any internal sequences that abut a <br> (\4) are replaced by \4
   --
   result = gsub_until_unchanged(result, '[\1\2\3\5\7\8]*\4+[\1\2\3\5\7\8]*', '\4')

   --
   -- Series of comma separated by space or dash or colon are replaced by one comma
   --
   result = gsub_until_unchanged(result, '\3[\2\1\5\7\8]*\3', '\3')

   --
   -- Series of colon separated by space or dash or comma are replaced by one comma
   --
   result = gsub_until_unchanged(result, '\5[\2\1\3\7\8]*\5', '\3')

   --
   -- Series of dashes separated by space or comma or colon are replaced by one dash
   --
   result = gsub_until_unchanged(result, '\1[\2\3\5\7\8]*\1', '\1')

   --
   -- Series of any of these touching \6 (a NOJOINERS token) are zapped.
   --
   result = gsub_until_unchanged(result, '[\1\2\3\4\5\6\7\8]*\6[\1\2\3\4\5\6\7\8]*', '')

   --
   -- Put everything else back, squeltching repeats
   --
   result = result:gsub('[\1\2\4\3\5\7\8]+', UnReplace)

   --DumpLog(result, "Spec after")

   return result, nil
end


-- =============================================================================================================================================



































function MetadataHandler.getLabelNumberFromColorName(colorName)
	if colorName == 'red' then
		return 1
	elseif colorName == 'yellow' then
		return 2
	elseif colorName == 'green' then
		return 3
	elseif colorName == 'blue' then
		return 4
	elseif colorName == 'purple' then
		return 5
	else
		return 0
	end
end

function MetadataHandler.createMetadataForPhoto(photo, outlets, labelMapToColorNames)

	local metadata,count = {},0

	local colorLabel = photo:getFormattedMetadata('label')
	if colorLabel ~= nil then
		count = count+1
		metadata['color-label'] = colorLabel
		
		local colorName = labelMapToColorNames[colorLabel]
		if colorName ~= nil then
			count = count+1
			metadata['color-name'] = colorName
			count = count+1
			metadata['color-number'] = MetadataHandler.getLabelNumberFromColorName(colorName)
		end
		
	end
	
	local gps = photo:getRawMetadata('gps');

	if gps ~= nil then
		metadata['latitude'] = gps.latitude
		metadata['longitude'] = gps.longitude
	end

	local title = photo:getFormattedMetadata('title')
	if title ~= nil and title ~= '' then
		count = count+1
		metadata['title'] = title
	end

	local caption = photo:getFormattedMetadata('caption')
	if caption ~= nil and caption ~= '' then
		count = count+1
		metadata['caption'] = caption
	end

	if outlets == nil then
		return metadata,count
	end
		
	for k,v in pairs(outlets) do
		count = count+1
		metadata[v.name] = MetadataHandler.forPhoto({}, v.value, {photo = photo})
	end

	return metadata,count
end

