local LrLogger = import 'LrLogger'
local logger = LrLogger("ttg")
logger:enable( "logfile" )

TTGGalleryIDHelper = {}

function compareAlbumTitles(a, b)
	
	if a == nil and b == nil then
		return no
	elseif a == nil then
		return yes
	elseif b == nil then
		return no
	elseif a.title == nil and b.title == nil then
		return no
	elseif a.title == nil then
		return yes
	elseif b.title == nil then
		return no
	else
		return a.title < b.title;
	end
end

function TTGGalleryIDHelper.new(o)
  o.service = o.service or nil
  o.galleriesXMLData = {}
  setmetatable(o, { __index = TTGGalleryIDHelper }) -- base table added to tell Lua what values are needed to look-up.
  return o
end

function TTGGalleryIDHelper:gatherIDsForCollection( collection )

	if  collection.getRemoteId ~= nil then

		local id =  "" 
		
		if id ~= collection:getRemoteId() then
			id =  collection:getRemoteId()
		end
		
		local slug = ""
		
		if collection:getRemoteUrl() ~= nil then
			local url =  collection:getRemoteUrl()
			urlC = TTGUtils.split(url, "/")
			slug = urlC[#urlC]
		end

		local parent = ""
	
		if 	collection:getParent() ~= nil then
			parent = collection:getParent():getRemoteId()
		end

		self.galleriesXMLData[#self.galleriesXMLData+1] = { id = id, slug = slug, parent = parent, type = "set" }
	end

	local childCollectionSets = collection:getChildCollectionSets()

	for i=1,#childCollectionSets,1 do
		self:gatherIDsForCollection(childCollectionSets[i])
	end

	local childCollections = collection:getChildCollections()
	for i=1,#childCollections,1 do
		local id =  "" 

		if childCollections[i]:getRemoteId() ~= nil then
			id = childCollections[i]:getRemoteId()
		end
		
		local slug = ""
		
		if childCollections[i]:getRemoteUrl() ~= nil then
			local url = childCollections[i]:getRemoteUrl()
			urlC = TTGUtils.split(url, "/")
			slug = urlC[#urlC]
		end
		
		local parent = ""
	
		if 	childCollections[i]:getParent() ~= nil then
			parent = childCollections[i]:getParent():getRemoteId()
		end

		self.galleriesXMLData[#self.galleriesXMLData+1] = { id = id, slug = slug, parent = parent, type = "album" }
	end
end

function TTGGalleryIDHelper:gatherAlbumSetsForCollection( collection )

	-- This is a Collection Set, and not the Service
	if  collection.getRemoteId ~= nil then 
		local info  = collection:getCollectionSetInfoSummary()
		local id = info.localIdentifier
	
		local collectionSettings = info.collectionSettings
		local title = collectionSettings.title
	
		self.galleriesXMLData[#self.galleriesXMLData+1] = { id = id, title = title }
	end
	
	local childCollectionSets = collection:getChildCollectionSets()

	for i=1,#childCollectionSets,1 do
		self:gatherAlbumSetsForCollection(childCollectionSets[i])
	end
end

function TTGGalleryIDHelper:gatherAlbumsForCollection( collection )
	local childCollectionSets = collection:getChildCollectionSets()

	for i=1,#childCollectionSets,1 do
		self:gatherAlbumsForCollection(childCollectionSets[i])
	end

	local childCollections = collection:getChildCollections()
	for i=1,#childCollections,1 do
		local info  = childCollections[i]:getCollectionInfoSummary()

		if info.localIdentifier ~= nil and info.collectionSettings.title ~= nil then
			self.galleriesXMLData[#self.galleriesXMLData+1] = { id = info.localIdentifier, title = info.collectionSettings.title }
		end

	end
end

function TTGGalleryIDHelper:findAlbumSetForServiceById( collection, album_id )

	-- This is a Collection Set, and not the Service
	if  collection.getRemoteId ~= nil then 
		local info  = collection:getCollectionSetInfoSummary()
		
		if info.localIdentifier == album_id then
			self.foundAlbumSet = info.collectionSettings
		end
	end
	
	local childCollectionSets = collection:getChildCollectionSets()

	for i=1,#childCollectionSets,1 do
		self:findAlbumSetForServiceById(childCollectionSets[i], album_id)
	end
end

function TTGGalleryIDHelper:findAlbumForServiceById( collection, album_id )
	
	local childCollectionSets = collection:getChildCollectionSets()

	for i=1,#childCollectionSets,1 do
		self:findAlbumForServiceById(childCollectionSets[i], album_id)
	end

	local childCollections = collection:getChildCollections()
	for i=1,#childCollections,1 do
		local info  = childCollections[i]:getCollectionInfoSummary()
		
		if info.localIdentifier == album_id then
			self.foundAlbum = info.collectionSettings
		end
	end
end

function TTGGalleryIDHelper:gatherIDsForPublisherService()
	if self.service then
		self:gatherIDsForCollection(self.service)
		return self.galleriesXMLData
	else
		return {}
	end
end

function TTGGalleryIDHelper:gatherAlbumSetsForPublisherService()
	if self.service then
		self:gatherAlbumSetsForCollection(self.service)
		table.sort(self.galleriesXMLData, compareAlbumTitles)
		return self.galleriesXMLData
	else
		return {}
	end
end

function TTGGalleryIDHelper:gatherAlbumsForPublisherService()
	if self.service then
		self:gatherAlbumsForCollection(self.service)
		table.sort(self.galleriesXMLData, compareAlbumTitles)
		return self.galleriesXMLData
	else
		return {}
	end
end

function TTGGalleryIDHelper:findAlbumSetById(album_id)
	self.foundAlbumSet = nil
	
	if self.service then
		self:findAlbumSetForServiceById(self.service, album_id)
	end
	
	return self.foundAlbumSet
end

function TTGGalleryIDHelper:findAlbumById(album_id)
	self.foundAlbum = nil
	
	if self.service then
		self:findAlbumForServiceById(self.service, album_id)
	end
	
	return self.foundAlbum
end

function TTGGalleryIDHelper:delimitedIDListForService()
	if self.service then
		self:gatherIDsForCollection(self.service)

		local ids = ''
		
		for i=1,#self.galleriesXMLData,1 do
			if i > 1 then
				ids = ids .. ','
			end
			
			ids = ids .. self.galleriesXMLData[i].id
		end
		
		return ids
	else
		return ''
	end
end