local LrLogger = import 'LrLogger'
local logger = LrLogger("ttg")
logger:enable( "logfile" )

TTGMigrator = {}

function TTGMigrator.new(o)
	if o.migrated_album_ids then
		o.migrated_album_ids = TTGUtils.set(o.migrated_album_ids)
		logger:debug('set: '..TTGUtils.inspect(o.migrated_album_ids))
	else
		o.migrated_album_ids = nil
	end

    o.source = o.source or nil
    o.destination = o.destination or nil
	o.migrated = {}
	o.omitted = {}

	setmetatable(o, { __index = TTGMigrator }) -- base table added to tell Lua what values are needed to look-up.
	return o
end

function TTGMigrator:getMigrated()
	return self.migrated
end

function TTGMigrator:getOmitted()
	return self.omitted
end

function TTGMigrator:copyCollection(collection, parent, service)
	local summary = collection:getCollectionInfoSummary()
	local settings = summary.collectionSettings

	local catalog = (import 'LrApplication').activeCatalog()
	local newCollection = service:createPublishedCollection(collection:getName(), parent, false)
	if newCollection ~= nil then
		local photos = collection:getPublishedPhotos()
		
		if photos ~= nil then
			newCollection:addPublishedPhotos(photos)
		end
		
		newCollection:setCollectionSettings(settings)
		newCollection:setRemoteId(collection:getRemoteId())			
		newCollection:setRemoteUrl(collection:getRemoteUrl())
		
		self.migrated[#self.migrated+1] = collection	
	end
end

function TTGMigrator:copyCollectionSet(collectionSet, parent, service)
	local summary = collectionSet:getCollectionSetInfoSummary()
	local settings = summary.collectionSettings

	local catalog = (import 'LrApplication').activeCatalog()
	local newCollectionSet = service:createPublishedCollectionSet(collectionSet:getName(), parent, false)
	
	if newCollectionSet ~= nil then
		newCollectionSet:setCollectionSetSettings(settings)
		newCollectionSet:setRemoteId(collectionSet:getRemoteId())
		newCollectionSet:setRemoteUrl(collectionSet:getRemoteUrl())
		
		self.migrated[#self.migrated+1] = collectionSet
	end

	local collectionSets = collectionSet:getChildCollectionSets()
	local collections = collectionSet:getChildCollections()
	
	for j=1,#collectionSets,1 do
		local childCollectionSet = collectionSets[j]

		if childCollectionSet:getRemoteId() and self.migrated_album_ids[""..childCollectionSet:getRemoteId()] then
			logger:debug("found (d): "..childCollectionSet:getRemoteId())
			self:copyCollectionSet(childCollectionSet, newCollectionSet, service)
		else
			logger:debug("NOT found (d): "..childCollectionSet:getRemoteId())
			self.omitted[#self.omitted+1] = childCollectionSet
			--self.migrated[#self.migrated+1] = childCollectionSet
		end
	end

	for j=1,#collections,1 do
		local childCollection = collections[j]
		
		if childCollection:getRemoteId() ~= nil then
			if self.migrated_album_ids[""..childCollection:getRemoteId()] then
				logger:debug("found (c): "..childCollection:getRemoteId())
				self:copyCollection(childCollection, newCollectionSet, service)
			else
				logger:debug("NOT found (c): "..childCollection:getRemoteId())
				self.omitted[#self.omitted+1] = childCollection
				--self.migrated[#self.migrated+1] = childCollection
			end
		else
			logger:debug("found NIL remoteId (c) for name:"..childCollection:getName())
		end
	end
end

function TTGMigrator:copyPublishService()
	local catalog = (import 'LrApplication').activeCatalog()
	
	catalog:withWriteAccessDo( "copyPublisherService", function( context ) 

		local collections = self.source:getChildCollections()
		local collectionSets = self.source:getChildCollectionSets()

		for j=1,#collections,1 do
			local collection = collections[j]

			logger:debug('set (b): '..TTGUtils.inspect(self.migrated_album_ids))
					
			if collection:getRemoteId() then
				if self.migrated_album_ids[""..collection:getRemoteId()] then
					logger:debug("found (b): "..collection:getRemoteId())
					self:copyCollection(collection, nil, self.destination)
				else
					logger:debug("NOT found (b): "..collection:getRemoteId())
					--local inspected = TTGUtils.inspect(collection)
					logger:debug('inspected: '..collection:getName())
					self.omitted[#self.omitted+1] = collection
					--self.migrated[#self.migrated+1] = collection
				end
			end
		end

		for j=1,#collectionSets,1 do
			local collectionSet = collectionSets[j]
			
			if self.migrated_album_ids[""..collectionSet:getRemoteId()] then
				logger:debug("found (a): "..collectionSet:getRemoteId())
				self:copyCollectionSet(collectionSet, nil, self.destination)
			else
				logger:debug("NOT found (a): "..collectionSet:getRemoteId())
				self.omitted[#self.omitted+1] = collectionSet
				--self.migrated[#self.migrated+1] = collectionSet
			end
		end
	end )
end	

function TTGMigrator.getServicesForPublisherId(publisherId, publisherName)
	local services = {}
	
	local catalog = (import 'LrApplication').activeCatalog()
	local publishServices = catalog:getPublishServices(publisherId)

	for i=1,#publishServices,1 do
		local service = publishServices[i]
		
		local serviceKV = {service = publishServices[i], value = #services+1}
		if publishServices[i]:getName() ~= publisherName then
			serviceKV.title = publisherName..": "..publishServices[i]:getName()
		else
			serviceKV.title = publishServices[i]:getName()
		end
		
		services[#services+1] = serviceKV		
	end
	
	return services
end

function TTGMigrator.getAllTTGServices()
	local services = {}
	
	local catalog = (import 'LrApplication').activeCatalog()

	local ce3Services =  catalog:getPublishServices("com.ttg.ce2.publisher")
	local ce4Services =  catalog:getPublishServices("com.ttg.ce4.publisher")

	for i=1,#ce3Services,1 do
		local serviceKV = {service = ce3Services[i], value = #services+1}
		if ce3Services[i]:getName() ~= "TTG CE3 Publisher" then
			serviceKV.title = "TTG CE3 Publisher: "..ce3Services[i]:getName()
		else
			serviceKV.title = ce3Services[i]:getName()
		end
		
		services[#services+1] = serviceKV		
	end

	for i=1,#ce4Services,1 do
		local serviceKV = {service = ce4Services[i], value = #services+1}
		if ce4Services[i].title ~= "TTG CE4 Publisher" then
			serviceKV.title = "TTG CE4 Publisher: "..ce4Services[i]:getName()
		else
			serviceKV.title = ce4Services[i]:getName()
		end
		
		services[#services+1] = serviceKV		
	end

	return services
end

function TTGMigrator.getServiceForPublisherIdAndName(serviceId, name)
	local catalog = (import 'LrApplication').activeCatalog()
	local publishServices = catalog:getPublishServices(serviceId)

	for i=1,#publishServices,1 do
		local service = publishServices[i]
		
		if service:getName() == name then
			return service
		end
	end
	
	return nil
end
	