require 'TTGGalleryIDHelper'
require 'TTGMigrator'
require 'TTGUtils'

local LrBinding = import 'LrBinding'
local LrDialogs = import 'LrDialogs'
local LrErrors = import 'LrErrors'
local LrFunctionContext = import 'LrFunctionContext' 
local LrTasks = import 'LrTasks'
local LrView = import 'LrView'

local bind = LrView.bind
local share = LrView.share

local LrLogger = import 'LrLogger'
local logger = LrLogger( 'ttg' )
logger:enable( "logfile" )

TTGOutputStructureMenuItem = {}

function TTGOutputStructureMenuItem.showServicesDialogue(services)
	return LrFunctionContext.callWithContext( 'migrate dialog', function( context )
		local key
		local url

		local f = LrView.osFactory()
		local properties = LrBinding.makePropertyTable( context )
	
		if #services > 0 then
			properties.service = services[1].value
		end

		local contents = f:column {
			bind_to_object = properties,
			spacing = f:control_spacing(),
			fill = 1,
      
			f:row {
				bind_to_object = collectionSettings,
				f:static_text {
					title = "Output the structure of Album Sets and Albums into the log file.",
					alignment = 'left'
				},
			},
		
			f:row {
				f:spacer {
					height = 5
				},
			},

			f:row {
				bind_to_object = properties,
				f:static_text {
					title = "Source",
					alignment = 'right',
					width = share 'title_width',
				},
				f:popup_menu {
						value = bind 'service',
						items = services,
						validate = function(view,value)
							end,
						fill_horizontal = 1
					},
				},
	    }
    
		while true do
			local result = LrDialogs.presentModalDialog {
				title = "Output Publisher Structure", 
				contents = contents,
			}

			if result == 'ok' then
				return properties.service
			else
				LrErrors.throwCanceled()	
			end
		end
		
	end)
end

function TTGOutputStructureMenuItem.output_structure()

	LrTasks.startAsyncTask( function()
		local services = TTGMigrator.getAllTTGServices()

		if #services ==  0 then
			LrDialogs.message("Error", "There are no publisher instances")
			return
		end

		local serviceIndex = TTGOutputStructureMenuItem.showServicesDialogue(services)
		--handle cancel
		if not serviceIndex then 
			return
		end
		
		logger:debug("serviceIndex: "..serviceIndex)
		
		local service = services[serviceIndex].service

		if service then
			logger:debug("service found: "..service:getName())
		end

		local idHelper = TTGGalleryIDHelper.new({service = service})
		local galleriesXMLData = idHelper:gatherIDsForPublisherService()
		local inspected = TTGUtils.inspect(galleriesXMLData)
		logger:debug('inspected: '..inspected)
		
		LrDialogs.message("Output completed for for "..service:getName(), inspected)
	end)
end

TTGOutputStructureMenuItem.output_structure()
